
import asyncio
import base64
import logging
import pickle
import redis
import rq.timeouts
import traceback

from rq import Connection
from rq.worker import Worker
from rq.connections import get_current_connection
from rq.queue import get_failed_queue

from colibris.taskqueue import UnpicklableException, TimeoutException
from colibris.taskqueue.base import TaskQueueBackend


DEFAULT_POLL_RESULTS_INTERVAL = 1  # seconds

logger = logging.getLogger(__name__)


class RQBackend(TaskQueueBackend):
    def __init__(self, host='localhost', port=6379, db=0, password=None,
                 poll_results_interval=DEFAULT_POLL_RESULTS_INTERVAL):

        self._host = host
        self._port = port
        self._db = db
        self._password = password
        self._poll_results_interval = poll_results_interval

        self._pending_results = []
        self._poll_loop_started = False
        self._connection = None
        self._queue = None

    def _get_connection(self):
        if self._connection is None:
            logger.debug('initializing redis connection')

            self._connection = redis.Redis(host=self._host, port=self._port, db=self._db, password=self._password)

        return self._connection

    def _get_queue(self):
        if self._queue is None:
            logger.debug('initializing task queue')

            self._queue = rq.Queue(connection=self._get_connection())

        return self._queue

    async def _poll_results(self):
        self._poll_loop_started = True
        logger.debug('starting results polling loop ')

        loop = asyncio.get_event_loop()

        while loop.is_running():
            remaining_results = []
            for tup in self._pending_results:
                try:
                    result, timeout, future = tup
                    result.refresh()

                    if result.is_finished and not future.cancelled():
                        future.set_result(result.return_value)

                    elif result.is_failed:
                        try:
                            exc_value = base64.b64decode(result.exc_info)
                            exc_value = pickle.loads(exc_value)

                        except Exception as e:
                            logger.error('failed to decode exception: %s', e, exc_info=True)
                            exc_value = UnpicklableException(result.exc_info)

                        # treat exceptions that could not be pickled
                        if isinstance(exc_value, str):
                            exc_value = UnpicklableException(exc_value)

                        # transform rq timeouts into taskqueue timeouts
                        if isinstance(exc_value, rq.timeouts.JobTimeoutException):
                            exc_value = TimeoutException(timeout)

                        future.set_exception(exc_value)

                    else:
                        remaining_results.append(tup)

                except Exception as e:
                    logger.error('polling results failed: %s', e, exc_info=True)

            self._pending_results = remaining_results

            await asyncio.sleep(self._poll_results_interval)

    async def execute(self, func, *args, timeout, **kwargs):
        if not self._poll_loop_started:
            _ = asyncio.ensure_future(self._poll_results())

        queue = self._get_queue()

        loop = asyncio.get_event_loop()
        future = loop.create_future()
        result = queue.enqueue(func, timeout=timeout, *args, **kwargs)

        self._pending_results.append((result, timeout, future))

        return await future

    @staticmethod
    def handle_task_exception(job, exc_type, exc_value, tb):
        try:
            pickled_exc_value = pickle.dumps(exc_value)

        except (pickle.PickleError, TypeError):
            logger.error('exception could not be pickled')
            exc_string = Worker._get_safe_exception_string(traceback.format_exception(exc_type, exc_value, tb))
            pickled_exc_value = pickle.dumps(exc_string)

        # we need to encode pickled_exc_value using base64, since pickle is not guaranteed to be ASCII
        pickled_exc_value = base64.b64encode(pickled_exc_value).decode()

        failed_queue = get_failed_queue(get_current_connection(), job.__class__)
        failed_queue.quarantine(job, exc_info=pickled_exc_value)

    def add_worker_arguments(self, parser):
        parser.add_argument('--queue-name', help='The task queue name ("default" if unspecified)',
                            type=str, default='default')

    def run_worker(self, options):
        with Connection(self._get_connection()):
            w = Worker([options.queue_name], exception_handlers=[self.handle_task_exception])
            w.work()




import logging
import os
import peewee

from peewee import *
try:
    from playhouse.postgres_ext import *
except ImportError:
    pass

from colibris import settings
from colibris import utils


_PEEWEE_DB_PARAMS_MAPPING = {
    'name': 'database',
    'username': 'user'
}

logger = logging.getLogger(__name__)
_database = None


def get_migrations_dir():

    # migrations live in the project root package
    return os.path.join(settings.PROJECT_PACKAGE_DIR, 'migrations')


def get_database():
    global _database

    if _database is None:
        backend_settings = dict(settings.DATABASE)
        backend_path = backend_settings.pop('backend', None)
        if not backend_path:
            return None

        backend_class = utils.import_member(backend_path)

        # translate settings into whatever peewee prefers
        for param, pparam in _PEEWEE_DB_PARAMS_MAPPING.items():
            if param in backend_settings:
                backend_settings[pparam] = backend_settings.pop(param)

        backend_settings.setdefault('autorollback', True)

        _database = backend_class(**backend_settings)
        _database.connect()

        logger.debug('db connection initialized')

    return _database


def connectivity_check():
    # run a dummy SQL statement to check connectivity with DB server

    try:
        list(get_database().execute_sql("select 'dummy'"))
        return True

    except Exception:
        return False


class Model(peewee.Model):
    # this is currently necessary for aiohttp-apispec request data validation
    def __iter__(self):
        return ((k, v) for (k, v) in self.__data__.items())

    def update_fields(self, fields):
        for n, v in fields.items():
            setattr(self, n, v)

    class Meta:
        database = get_database()




from colibris import app
from colibris import persist

#
# def init(webapp, loop):
#     pass  # add your coroutines to the loop here
#
#
# def get_health():
#     # return your health status or raise app.HealthException()
#
#     if not persist.connectivity_check():
#         raise app.HealthException('database connectivity check failed')
#
#     return 'healthy'
#




import asyncio
import inspect
import logging
import time

from aiohttp import web, hdrs
from aiohttp_apispec import setup_aiohttp_apispec
from aiohttp_swagger import setup_swagger

from colibris import routes as default_routes
from colibris import settings
from colibris import utils


logger = logging.getLogger(__name__)
middleware = []
route_auth_mapping = {}

_project_app = None
_start_time = time.time()


# webapp & middleware

def _init_webapp():
    for path in settings.MIDDLEWARE:
        middleware.append(utils.import_member(path))

    return web.Application(middlewares=middleware, client_max_size=settings.MAX_REQUEST_BODY_SIZE)


# app

class HealthException(Exception):
    pass


async def _init_app(app):
    global _project_app

    _project_app = utils.import_module_or_none('{}.app'.format(settings.PROJECT_PACKAGE_NAME))
    if _project_app is not None:
        init = getattr(_project_app, 'init', None)
        if init:
            init(app, asyncio.get_event_loop())


async def get_health():
    if hasattr(_project_app, 'get_health'):
        gh = _project_app.get_health
        if inspect.iscoroutinefunction(gh):
            h = await gh()

        else:
            h = gh()

        return h

    return {
        'uptime': int(time.time() - _start_time)
    }


async def _initial_health_check(app):
    try:
        await get_health()

    except Exception as e:
        logger.critical('initial health check problem: %s', e)
        raise


# routes

def _add_route_tuple(route):
    while len(route) < 4:
        route = route + (None,)

    method, path, handler, authorize = route

    if inspect.isclass(handler):  # class-based view
        method = hdrs.METH_ANY

    # aiohttp reuses the last resource if and only if two successive routes are the same,
    # which is kind of random.

    # Explicitly add resource instead of using add_route(),
    # since we want to prevent reusing the last resource.

    resource = webapp.router.add_resource(path)
    resource_route = resource.add_route(method, handler)

    route_auth_mapping[resource_route] = authorize


def _add_static_route_tuple(route):
    fs_path, prefix = route

    webapp.router.add_static(prefix, fs_path)


def _init_default_routes():
    for route in default_routes.ROUTES:
        _add_route_tuple(route)

    for route in default_routes.STATIC_ROUTES:
        _add_static_route_tuple(route)


def _init_project_routes():
    project_routes = utils.import_module_or_none('{}.routes'.format(settings.PROJECT_PACKAGE_NAME))
    if project_routes is None:
        return

    for _route in getattr(project_routes, 'ROUTES', []):
        _add_route_tuple(_route)

    for _route in getattr(project_routes, 'STATIC_ROUTES', []):
        _add_static_route_tuple(_route)


# apispec/swagger support

def _init_swagger():
    async def init_wrapper(app):
        setup_swagger(app=app, swagger_url=settings.API_DOCS_PATH, swagger_info=app['swagger_dict'])

    setup_aiohttp_apispec(app=webapp, title='API Documentation')
    webapp.on_startup.append(init_wrapper)


webapp = _init_webapp()

_init_project_routes()
_init_default_routes()
_init_swagger()

webapp.on_startup.append(_init_app)
webapp.on_startup.append(_initial_health_check)




from colibris.authorization import ANY_PERMISSION
from colibris.authorization.model import ModelBackend


class RightsBackend(ModelBackend):
    def __init__(self, resource_field, operations_field, **kwargs):
        self.resource_field = resource_field
        self.operations_field = operations_field

        super().__init__(**kwargs)

    def get_resource_field(self):
        return getattr(self.model, self.resource_field)

    def get_operations(self, right):
        return getattr(right, self.operations_field)

    def authorize(self, account, method, path, required_permissions):
        if required_permissions == ANY_PERMISSION:
            return True

        resource, required_operations = required_permissions.split(':', 1)

        query = (self.get_account_field() == account and
                 self.get_resource_field() == resource)

        # gather all rights from all entries for the given account and resource;
        allowed_operations = set()
        rights = self.model.select().where(query)
        for right in rights:
            allowed_operations |= set(self.get_operations(right))

        # consider the request authorized if all required operations are included in allowed operations
        required_operations = set(required_operations)
        return len(required_operations - allowed_operations) == 0




# Patch marshmallow to provide marshmallow.compat.string_types,
# so that marshmallow-peewee<=2.2.0 can be used.
# This can be removed as soon as the following PR gets merged:
# https://github.com/klen/marshmallow-peewee/pull/61/commits/462b78aea70ff818ab501be2d3d2cead420c274e

import marshmallow.compat
marshmallow.compat.string_types = (str,)





from aiohttp import web
from aiohttp_apispec import docs, request_schema, response_schema

from colibris import api
from colibris.schemas import many_envelope
from colibris.shortcuts import get_object_or_404

from __packagename__ import models
from __packagename__ import schemas

#
# View examples:
#
#
# @docs(tags=['Users'],
#       summary='Reveal details about the current user')
# @response_schema(schemas.UserSchema())
# def get_me(request):
#     if request.account:
#         result = schemas.UserSchema().dump(request.account)
#
#     else:
#         result = None
#
#     return web.json_response(result)
#
#
# @docs(tags=['Users'],
#       summary='Reveal details about a specific user')
# @response_schema(schemas.UserSchema())
# def get_user(request):
#     user_id = request.match_info['id']
#     user = get_object_or_404(models.User, user_id)
#     result = schemas.UserSchema().dump(user)
#
#     return web.json_response(result)
#
#
# @docs(tags=['Users'],
#       summary='List all users')
# @response_schema(many_envelope(schemas.UserSchema))
# def list_users(request):
#     users = models.User.select().order_by(models.User.username.asc())
#     result = schemas.UserSchema(many=True).dump(list(users))
#
#     return web.json_response(result)
#
#
# @docs(tags=['Users'],
#       summary='Add a new user')
# @request_schema(schemas.UserSchema())
# @response_schema(schemas.UserSchema())
# def add_user(request):
#     data = request['data']
#
#     if models.User.select().where(models.User.username == data['username']).exists():
#         raise api.DuplicateException(models.User, 'username')
#
#     user = models.User.create(**data)
#     result = schemas.UserSchema().dump(user)
#
#     return web.json_response(result, status=201)
#
#
# @docs(tags=['Users'],
#       summary='Update an existing user')
# @request_schema(schemas.UserSchema(partial=True))
# @response_schema(schemas.UserSchema(partial=True))
# def update_user(request):
#     user_id = request.match_info['id']
#     user = get_object_or_404(models.User, user_id)
#     data = request['data']
#
#     if 'username' in data:
#         query = (models.User.username == data['username'] and models.User.id != user_id)
#         if models.User.select().where(query).exists():
#             raise api.DuplicateException(models.User, 'username')
#
#     user.update_fields(data)
#
#     user.save()
#     result = schemas.UserSchema().dump(user)
#
#     return web.json_response(result)
#
#
# @docs(tags=['Users'],
#       summary='Delete a user')
# def delete_user(request):
#     user_id = request.match_info['id']
#     if models.User.delete().where(models.User.id == user_id).execute() == 0:
#         raise api.ModelNotFoundException(models.User)
#
#     return web.json_response(status=204)
#




from peewee_migrate import Router

from colibris import persist

from . import BaseCommand


class MigrateCommand(BaseCommand):
    def execute(self, options):
        router = Router(persist.get_database(), migrate_dir=persist.get_migrations_dir())
        router.run()




import argparse
import logging.config
import sys

from colibris import settings
from colibris import utils


class BaseCommand:
    def __init__(self, args):
        self.args = args
        self.parser = argparse.ArgumentParser()
        self.add_arguments(self.parser)

    def initialize(self):
        pass

    def run(self):
        self.initialize()
        options = self.parser.parse_args(self.args)
        self.execute(options)

    def add_arguments(self, parser):
        pass

    def execute(self, options):
        raise NotImplementedError

    @classmethod
    def get_name(cls):
        return cls.__name__[0:-7].lower()


def gather_all_commands():
    global ALL_COMMANDS

    # import any project-specific commands
    utils.import_module_or_none('{}.commands'.format(settings.PROJECT_PACKAGE_NAME))

    ALL_COMMANDS = {c.get_name(): c for c in BaseCommand.__subclasses__()}


def show_commands_usage():
    cmds = '\n  '.join([c.get_name() for c in ALL_COMMANDS.values()])
    usage = '{prog} <command> [args]\n\n' \
            'available commands:\n  {cmds}\n\n' \
            'use {prog} <command> -h for help'

    usage = usage.format(cmds=cmds, prog='manage.py')

    print(usage)


def main():
    gather_all_commands()

    args = sys.argv
    if len(args) < 2 or args[1] not in ALL_COMMANDS:
        show_commands_usage()
        sys.exit(1)

    # configure logging
    logging_config = dict(settings.LOGGING)
    logging_config['disable_existing_loggers'] = False
    utils.dict_update_rec(logging_config, settings.LOGGING_OVERRIDES)

    logging.config.dictConfig(logging_config)

    command_class = ALL_COMMANDS[sys.argv[1]]
    command = command_class(args[2:])
    command.run()


from . import makemigrations
from . import migrate
from . import runserver
from . import runworker
from . import shell




import jwt
import re

from colibris.authentication.exceptions import AuthenticationException
from colibris.authentication.model import ModelBackend


_AUTH_HEADER = 'Authorization'
_AUTH_TOKEN_REGEX = re.compile('Bearer (.+)', re.IGNORECASE)


class JWTException(AuthenticationException):
    pass


class JWTBackend(ModelBackend):
    def __init__(self, identity_claim='sub', **kwargs):
        self.identity_claim = identity_claim

        super().__init__(**kwargs)

    def extract_auth_data(self, request):
        auth_header = request.headers.get(_AUTH_HEADER)
        if auth_header is None:
            raise JWTException('missing authorization header')

        m = _AUTH_TOKEN_REGEX.match(auth_header)
        if not m:
            raise JWTException('invalid authorization header')

        token = m.group(1)

        try:
            jwt_decoded = jwt.decode(token, verify=False)

        except jwt.DecodeError:
            raise JWTException('invalid authorization header')

        identity = jwt_decoded.get(self.identity_claim)
        if identity is None:
            raise JWTException('missing identity claim')

        return identity, token

    def verify_identity(self, account, auth_data):
        secret = self.get_secret(account)

        try:
            jwt.decode(auth_data, key=secret, verify=True)

        except jwt.InvalidSignatureError:
            return False

        return True




DEBUG = True

LISTEN = '0.0.0.0'
PORT = 8888

# DATABASE = {
#     'backend': 'colibris.persist.PostgresqlDatabase',
#     'name': '__packagename__',
#     'host': '127.0.0.1',
#     'port': 5432,
#     'username': 'username',
#     'password': 'password'
# }

# AUTHENTICATION = {
#     'backend': 'colibris.authentication.jwt.JWTBackend',
#     'model': '__packagename__.models.User',
#     'identity_claim': 'sub',
#     'identity_field': 'username',
#     'secret_field': 'password'
# }

# AUTHORIZATION = {
#     'backend': 'colibris.authorization.role.RoleBackend',
#     'role_field': 'role'
# }




import async_timeout
import asyncio
import logging

from colibris import settings
from colibris import utils

from .exceptions import *


DEFAULT_TIMEOUT = 300  # seconds

logger = logging.getLogger(__name__)

_backend = None


def get_backend():
    global _backend

    if _backend is None:
        backend_settings = dict(settings.TASK_QUEUE)
        backend_path = backend_settings.pop('backend', None)
        if backend_path is None:
            raise TaskQueueNotConfigured()

        backend_class = utils.import_member(backend_path)
        _backend = backend_class(**backend_settings)

        logger.debug('backend of class %s created', backend_path)

    return _backend


async def execute(func, *args, timeout=DEFAULT_TIMEOUT, **kwargs):
    try:
        with async_timeout.timeout(timeout):
            return await get_backend().execute(func, *args, timeout=timeout, **kwargs)

    except asyncio.TimeoutError:
        raise TimeoutException(timeout)


def add_worker_arguments(parser):
    get_backend().add_worker_arguments(parser)


def run_worker(options):
    get_backend().run_worker(options)
